# Generated by Django 3.1.8 on 2021-04-07 14:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0009_auto_20210330_1659"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataset",
            name="schema_data",
            field=models.JSONField(verbose_name="Amsterdam Schema Contents"),
        ),
        migrations.AlterField(
            model_name="profile",
            name="schema_data",
            field=models.JSONField(verbose_name="Amsterdam Schema Contents"),
        ),
    ]
