# Generated by Django 4.1.7 on 2024-01-10 16:59

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("datasets", "0016_dataset_view_data"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="enable_export",
            field=models.BooleanField(default=False),
        ),
    ]
