# Generated by Django 3.0.3 on 2020-02-26 07:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0003_datasettable_geometry_field_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="auth",
            field=models.CharField(
                blank=True, max_length=250, null=True, verbose_name="Authorization"
            ),
        ),
        migrations.AddField(
            model_name="datasettable",
            name="auth",
            field=models.CharField(blank=True, max_length=250, null=True),
        ),
    ]
